import { Logger } from '@nestjs/common';
import {
  ConnectedSocket,
  MessageBody,
  OnGatewayConnection,
  OnGatewayDisconnect,
  OnGatewayInit,
  SubscribeMessage,
  WebSocketGateway,
  WebSocketServer,
} from '@nestjs/websockets';
import { Request } from 'express';
import { Packet } from 'src/net/packet';
import { PlayerStoreService } from 'src/player/player-store.service';
import { PlayerSocket } from 'src/types/data-socket';
import { ErrorType } from 'src/types/error-type.enum';
import { PacketType } from 'src/types/packet-type.enum';
import { getRandomId } from 'src/utils/random';
import { WorldService } from 'src/world/world.service';
import { Server } from 'ws';

@WebSocketGateway()
export class EventsGateway
  implements OnGatewayInit, OnGatewayConnection, OnGatewayDisconnect
{
  private readonly logger = new Logger(EventsGateway.name);

  constructor(
    private readonly players: PlayerStoreService,
    private readonly world: WorldService,
  ) {}

  @WebSocketServer()
  server: Server;

  @SubscribeMessage('message')
  async onEvent(
    @ConnectedSocket() client: PlayerSocket,
    @MessageBody() data: Packet,
  ) {
    let player = this.players.getByClient(client);
    if (!player && data.packet !== PacketType.AUTH) {
      return client.close(
        1007,
        new Packet(PacketType.ERROR)
          .write(ErrorType.AUTH_FAIL, 'uint8')
          .toBuffer(),
      );
    }

    if (data.packet === PacketType.AUTH) {
      try {
        player = await this.players.authenticate(client, data);
      } catch {
        return client.close(
          1007,
          new Packet(PacketType.ERROR)
            .write(ErrorType.AUTH_FAIL, 'uint8')
            .toBuffer(),
        );
      }
      this.logger.log(
        `Client ${client.id} (${client.ip}) has logged in as ${player.name} (${player.id})`,
      );
      await this.world.initializePlayer(player);
    }

    if (this.world.handlePlayerPacket(data, player)) return false;

    return false;
  }

  afterInit(server: Server) {
    this.logger.log(
      `>>> Game server is listening on port ${server.options.port} <<<`,
    );
  }

  handleConnection(client: PlayerSocket, request: Request) {
    client.id = getRandomId();
    client.ip = request.socket.remoteAddress;
    this.logger.log(`Client ${client.id} connected from ${client.ip}.`);
  }

  handleDisconnect(client: PlayerSocket) {
    this.logger.log(`Client  ${client.id} (${client.ip}) disconnected.`);
    this.world.playerQuit(client);
  }
}
